from typing import Dict

from openapi_core import validate_request
from openapi_core.templating.security.exceptions import SecurityNotFound
from openapi_core.validation.request.exceptions import ParameterValidationError
from openapi_core.validation.request.validators import APICallRequestValidator
from openapi_core.validation.schemas.exceptions import InvalidSchemaValue

from powertools_oas_validator.exceptions import UnsupportedOpenAPIVersion
from powertools_oas_validator.overrides import (
    V30RequestUnmarshaller,
    V31RequestUnmarshaller,
)
from powertools_oas_validator.services.error_handler import ErrorHandler
from powertools_oas_validator.services.event_parser import EventParserProtocol
from powertools_oas_validator.services.spec_loader import SpecLoaderProtocol
from powertools_oas_validator.services.spec_parser import SpecParser

marshaller_map = {"3.1": V31RequestUnmarshaller, "3.0": V30RequestUnmarshaller}


class SpecValidator:
    def __init__(
        self,
        file_path: str,
        event: Dict,
        spec_loader: SpecLoaderProtocol,
        event_parser: EventParserProtocol,
    ) -> None:
        self.file_path = file_path
        self.event = event
        self.spec_loader = spec_loader
        self.event_parser = event_parser

    def validate_request_against_spec(self) -> None:
        self.spec = self.spec_loader.read_from_file_name()

        request = self.event_parser.event_to_request()

        try:
            validate_request(
                request,
                spec=self.spec,
                base_url=request.host_url,
                cls=self._get_class(),  # type: ignore
            )

        except (ParameterValidationError, InvalidSchemaValue, SecurityNotFound) as ex:
            ErrorHandler.raise_schema_validation_error(ex, request)

    def _get_class(self) -> type[APICallRequestValidator]:
        version = SpecParser.get_openapi_version(self.spec)

        try:
            return marshaller_map[f"{version.major}.{version.minor}"]
        except KeyError:
            raise UnsupportedOpenAPIVersion(
                f"Unsupported OpenAPI version: '{str(version)}'"
            )



from openapi_core import V30RequestValidator, V31RequestValidator
from openapi_core.unmarshalling.request.unmarshallers import BaseRequestUnmarshaller
from openapi_core.unmarshalling.schemas import (
    oas30_write_schema_unmarshallers_factory,
    oas31_schema_unmarshallers_factory,
)
from openapi_core.unmarshalling.unmarshallers import BaseUnmarshaller
from openapi_core.validation.request.validators import APICallRequestValidator

from powertools_oas_validator.overrides.validators import CustomBaseRequestValidator


class CustomBaseRequestUnmarshaller(
    BaseRequestUnmarshaller, CustomBaseRequestValidator, BaseUnmarshaller
):
    ...


class BaseAPICallRequestUnmarshaller(CustomBaseRequestUnmarshaller):
    pass


class APICallRequestUnmarshaller(
    APICallRequestValidator, BaseAPICallRequestUnmarshaller
):
    ...


#
class V30RequestUnmarshaller(V30RequestValidator, APICallRequestUnmarshaller):
    schema_unmarshallers_factory = oas30_write_schema_unmarshallers_factory


class V31RequestUnmarshaller(V31RequestValidator, APICallRequestUnmarshaller):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory



import re
from typing import List, Union

from aws_lambda_powertools.utilities.validation.exceptions import SchemaValidationError
from openapi_core.templating.security.exceptions import SecurityNotFound
from openapi_core.validation.request.exceptions import ParameterValidationError
from openapi_core.validation.schemas.exceptions import InvalidSchemaValue

from powertools_oas_validator.types import Request


class ErrorHandler:
    @staticmethod
    def raise_schema_validation_error(
        ex: Union[ParameterValidationError, InvalidSchemaValue, SecurityNotFound],
        request: Request,
    ) -> None:
        ex_type = type(ex)

        if issubclass(ex_type, ParameterValidationError):
            error = ErrorHandler._handle_parameter_error(ex, request)  # type: ignore
        elif ex_type == InvalidSchemaValue:
            error = ErrorHandler._handle_body_error(ex, request)  # type: ignore
        elif ex_type == SecurityNotFound:
            error = ErrorHandler._handle_security_error(ex, request)  # type: ignore
        else:
            raise ValueError(
                f"'{ex_type}' is not mapped to a function. Consider adding it."
            )
        raise error

    @staticmethod
    def _handle_security_error(
        ex: SecurityNotFound, request: Request
    ) -> SchemaValidationError:
        violating_schemes = ex.schemes[0]
        name = ErrorHandler._get_name(request, "security", violating_schemes[0])

        validation_message = f"'{ex.schemes}' are required security scheme(s)."

        return SchemaValidationError(
            message=None,
            validation_message=validation_message,
            name=name,
            path=ErrorHandler._get_path(name),
            value=None,
            definition=None,
            rule=None,
            rule_definition=None,
        )

    @staticmethod
    def _handle_parameter_error(
        ex: ParameterValidationError, request: Request
    ) -> SchemaValidationError:
        name = ErrorHandler._get_name(request, "parameters", ex.name)  # type: ignore

        validation_message = f"'{ex.name}' is a required {ex.location} parameter."

        return SchemaValidationError(
            message=None,
            validation_message=validation_message,
            name=name,
            path=ErrorHandler._get_path(name),
            value=None,
            definition=None,
            rule=None,
            rule_definition=None,
        )

    @staticmethod
    def _handle_body_error(
        ex: InvalidSchemaValue, request: Request
    ) -> SchemaValidationError:
        violating_req_params = []

        for error in ex.schema_errors:
            if "required" in error.message:  # type: ignore
                try:
                    violating_req_params.append(
                        re.search("'(.+?)'", error.message).group(1)  # type: ignore
                    )
                except IndexError:
                    ...
            else:
                raise ValueError(
                    (
                        "Unmapped error type. Consider adding case for"
                        + f"'{ex.message}'"  # type: ignore
                    )
                )

        try:
            name = ErrorHandler._get_name(
                request, "requestBody", violating_req_params[0]
            )
            path = ErrorHandler._get_path(name)

        except KeyError:
            name = ErrorHandler._get_name(request, "request_body", "")
            path = ErrorHandler._get_path("")

        validation_message = ""

        if violating_req_params:
            try:
                validation_message = f"{violating_req_params} are required propertie(s)"
            except KeyError:
                ...

        return SchemaValidationError(
            message=None,
            validation_message=validation_message,
            name=name,
            path=path,
            value=None,
            definition=None,
            rule=None,
            rule_definition=None,
        )

    @staticmethod
    def _get_name(request: Request, path: str, violating_param: str) -> str:
        return (
            request.path.replace("/", ".").lstrip(".") + f".{path}[{violating_param}]"
        )

    @staticmethod
    def _get_path(name: str) -> List[str]:
        return name.replace("[", ".").replace("]", "").split(".")


error_to_func = {
    ParameterValidationError: ErrorHandler._handle_parameter_error,
    InvalidSchemaValue: ErrorHandler._handle_body_error,
}



from typing import Any, Optional

from openapi_core import Spec
from openapi_core.datatypes import RequestParameters
from openapi_core.exceptions import OpenAPIError
from openapi_core.validation.decorators import ValidationErrorWrapper
from openapi_core.validation.request.validators import BaseRequestValidator, Dict


class CustomBaseRequestValidator(BaseRequestValidator):
    class CustomValidationErrorWrapper(ValidationErrorWrapper):
        ...

    @CustomValidationErrorWrapper(OpenAPIError)
    def _get_body(self, body: Optional[str], mimetype: str, operation: Spec) -> Any:
        return super()._get_body.__wrapped__(  # type: ignore
            self, body=body, mimetype=mimetype, operation=operation
        )

    @CustomValidationErrorWrapper(OpenAPIError)
    def _get_security(
        self, parameters: RequestParameters, operation: Spec
    ) -> Optional[Dict[str, str]]:
        return super()._get_security.__wrapped__(  # type: ignore
            self, parameters=parameters, operation=operation
        )



from collections.abc import Callable
from typing import Dict

from aws_lambda_powertools.middleware_factory import lambda_handler_decorator
from aws_lambda_powertools.utilities.typing import LambdaContext

from powertools_oas_validator.services.event_parser import EventParser
from powertools_oas_validator.services.spec_loader import SpecLoader
from powertools_oas_validator.services.spec_validator import SpecValidator


@lambda_handler_decorator
def validate_request(
    handler: Callable,
    event: Dict,
    context: LambdaContext,
    oas_path: str,
) -> Callable:
    # Setup Spec Loader
    spec_loader = SpecLoader(oas_path)

    # Setup Event Parser
    event_parser = EventParser(event)

    # Setup SpecValidator
    spec_validator = SpecValidator(oas_path, event, spec_loader, event_parser)

    # Validate Spec Against Event
    spec_validator.validate_request_against_spec()

    return handler(event, context)



import json
from typing import Dict, Protocol

from openapi_core.datatypes import RequestParameters
from werkzeug.datastructures import Headers, ImmutableMultiDict

from powertools_oas_validator.exceptions import InvalidEventError
from powertools_oas_validator.types import Request


class EventParserProtocol(Protocol):
    def event_to_request(self) -> Request:
        raise NotImplementedError  # pragma: nocover


class EventParser(EventParserProtocol):
    def __init__(self, event: Dict) -> None:
        self.event = event

    def event_to_request(self) -> Request:
        return Request(
            host_url=self.get_host_url(),
            path=self.get_path(),
            full_url_pattern=self.get_full_url_pattern(),
            method=self.get_method(),
            parameters=self.get_parameters(),
            body=self.get_body(),
            mimetype=self.get_mimetype(),
        )

    def _get_headers(self) -> Dict:
        try:
            return self.event["headers"]
        except KeyError:
            raise InvalidEventError("'headers' missing from event.")

    def get_path(self) -> str:
        try:
            return self.event["path"]
        except KeyError:
            raise InvalidEventError("'path' missing from event.")

    def get_host_url(self) -> str:
        try:
            proto = self._get_headers()["X-Forwarded-Proto"]
        except KeyError:
            raise InvalidEventError("'headers.X-Forwarded-Proto' missing from event.")

        try:
            host = self._get_headers()["Host"]
        except KeyError:
            raise InvalidEventError("'headers.Host' missing from event.")

        return f"{proto}://{host}"

    def get_method(self) -> str:
        try:
            return self.event["httpMethod"].lower()
        except KeyError:
            raise InvalidEventError("'httpMethod' missing from event.")

    def get_mimetype(self) -> str:
        try:
            return self._get_headers()["Content-Type"]
        except KeyError:
            raise InvalidEventError("'headers.Content-Type' missing from event.")

    def get_full_url_pattern(self) -> str:
        host_url = self.get_host_url()

        try:
            return host_url + self.event["path"]
        except KeyError:
            raise InvalidEventError("'path' missing from event.")

    def get_parameters(self) -> RequestParameters:
        try:
            query_params = self.event["queryStringParameters"]
        except KeyError:
            query_params = {}

        query_params = ImmutableMultiDict(query_params)

        headers = Headers(self._get_headers())

        cookies: Dict = {}  # TODO: support cookies

        cookies = ImmutableMultiDict(cookies)

        try:
            path_params = self.event["pathParameters"]
        except KeyError:
            path_params = {}

        return RequestParameters(
            query=query_params, header=headers, cookie=cookies, path=path_params
        )

    def get_body(self) -> str:
        try:
            body = self.event["body"]
        except KeyError:
            return ""

        if type(body) == str:
            return body
        else:
            return json.dumps(body)



from dataclasses import dataclass

from openapi_core.datatypes import RequestParameters
from openapi_core.protocols import Request as CoreRequest


@dataclass
class OpenAPIVersion:
    major: int
    minor: int
    pico: int

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.pico}"


class Request(CoreRequest):
    def __init__(
        self,
        host_url: str,
        path: str,
        full_url_pattern: str,
        method: str,
        parameters: RequestParameters,
        body: str,
        mimetype: str,
    ) -> None:
        self._host_url = host_url
        self._path = path
        self._full_url_pattern = full_url_pattern
        self._method = method
        self.parameters = parameters
        self._body = body
        self._mimetype = mimetype

    @property
    def host_url(self) -> str:
        return self._host_url

    @host_url.setter
    def host_url(self, val: str) -> None:
        self._host_url = val

    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, val: str) -> None:
        self._path = val

    @property
    def full_url_pattern(self) -> str:
        return self._full_url_pattern

    @full_url_pattern.setter
    def full_url_pattern(self, val: str) -> None:
        self._full_url_pattern = val

    @property
    def method(self) -> str:
        return self._method

    @method.setter
    def method(self, val: str) -> None:
        self._method = val

    @property
    def body(self) -> str:
        return self._body

    @body.setter
    def body(self, val: str) -> None:
        self._body = val

    @property
    def mimetype(self) -> str:
        return self._mimetype

    @mimetype.setter
    def mimetype(self, val: str) -> None:
        self._mimetype = val



import os
from typing import Protocol

from openapi_core import Spec

from powertools_oas_validator.exceptions import (
    FileNotExistsError,
    NotSupportedFileTypeError,
)


class SpecLoaderProtocol(Protocol):
    def read_from_file_name(self) -> Spec:
        raise NotImplementedError  # pragma: nocover


class SpecLoader:
    oas_path: str
    validated_cache: bool = False

    def __init__(self, oas_path: str) -> None:
        self.oas_path = oas_path

    def read_from_file_name(self) -> Spec:
        if not self.validated_cache:
            self.validate_file()

        return Spec.from_file_path(self.oas_path)

    def validate_file(self) -> None:
        if self.validated_cache:
            return

        if not os.path.isfile(self.oas_path):
            raise FileNotExistsError(f"File does not exist on path: '{self.oas_path}'")

        extension = self.oas_path.split(".")[-1]

        if extension not in ["yaml", "json"]:
            raise NotSupportedFileTypeError(
                f"'.{extension} not supported. Only '.json' and '.yaml'"
            )

        self.validated_cache = True



from openapi_core import Spec

from powertools_oas_validator.types import OpenAPIVersion


class SpecParser:
    @staticmethod
    def get_openapi_version(spec: Spec) -> OpenAPIVersion:
        openapi_version = spec.accessor.lookup["openapi"]  # type: ignore

        openapi_version = openapi_version.split(".")

        return OpenAPIVersion(
            openapi_version[0], openapi_version[1], openapi_version[2]
        )
